// -*-C++-*- 

// Copyright (C) 2004 
// Christian Stimming <stimming@tuhh.de>

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2, or (at
// your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License along
// with this library; see the file COPYING.  If not, write to the Free
// Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.

//      LAPACK++ (V. 1.1)
//      (C) 1992-1996 All Rights Reserved.

/** @file
 * @brief Matrix index class
 */

#ifndef _LA_INDEX_H_
#define _LA_INDEX_H_

#include <iostream>   
#include <stdio.h>

/** @brief Matrix index class.
 *
 * This index class is used to adress submatrices of other matrices,
 * without actually copying any matrix elements in memory.
 *
 *  Note that we name it "LaIndex" to avoid confusion
 * with the "index()" string function in C, or other generic Index()
 * functions.
 */
class LaIndex
{
   private:
      int start_;
      int inc_;
      int end_;
   public:
      /** Construct a null index. Start, Increment, and End are
       * zero. null() will return true. */
      inline LaIndex() { start_ = inc_ = end_ = 0;}
      
      /** Construct an index that points only to one element, specified
       * by the given argument.
       * 
       * Start and End is the given argument; Increment is unity but has
       * no effect. */
      inline LaIndex(int start) 
	 : start_(start)
	 , inc_(1)
	 , end_(start)
      {}

      /** Construct an index that points to a range of elements,
       * starting from the first argument and ending at the second
       * argument.
       *
       * Start and End are given by the two arguments. Increment is
       * unity. */
      inline LaIndex(int start, int end)
	 : start_(start)
	 , inc_(1)
	 , end_(end)
      {}

      /** Construct an index that points to a range of elements with a
       * given increment.
       *
       * Start, End, and Increment are given by the three
       * arguments. */
      inline LaIndex(int start, int end, int increment)
	 : start_(start)
	 , inc_(increment)
	 , end_(end)
      {}

      /** Copy constructor */
      inline LaIndex(const LaIndex &s)
	 : start_(s.start_)
	 , inc_(s.inc_)
	 , end_(s.end_)
      {}

// ** must have multply defined start(), inc() and end() member functions
// ** for both const and non-const objects because compiler complains in 
// ** LaVector*.h, for example, about assignment to const member. 
// ** (LaVector*.h line 112, 113, 114)

        inline int& start() { return start_;}
        inline const int& start() const { return start_;}
        inline int& inc() { return inc_;}
        inline const int& inc() const { return inc_;}
        inline int& end() { return end_;}
        inline const int& end() const { return end_;}
        inline int length() const { return ((end()-start())/inc() + 1);}
        inline const int null() const{ return (start() == 0 && 
            inc() == 0  && end() == 0);}
        inline LaIndex& operator()(int start, int end){
            start_=start; inc_=1; end_=end; return *this;}
        inline LaIndex& operator+(int i){
            start_+=i; end_+=i; return *this;}
        inline LaIndex& operator=(const LaIndex& i){
            start_=i.start_; inc_=1; end_=i.end_; 
            return *this;}
};

inline std::ostream& operator<<(std::ostream& s, const LaIndex i)
{
    s << "(" << i.start() << ":" << i.inc() << ":" << i.end() << ")";

    return s;
}

#endif  
//  LA_INDEX_H_

