// -*-C++-*- 

// Copyright (C) 2004 
// Christian Stimming <stimming@tuhh.de>

// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2, or (at
// your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License along
// with this library; see the file COPYING.  If not, write to the Free
// Software Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.

//      LAPACK++ (V. 1.1)
//      (C) 1992-1996 All Rights Reserved.

/** @file
 * @brief Functions for solving linear equations
 */

#ifndef _LASLV_H
#define _LASLV_H

#include "lafnames.h"
#include LA_GEN_MAT_DOUBLE_H
#ifdef LA_COMPLEX_SUPPORT
# include LA_GEN_MAT_COMPLEX_H
#endif
#include LA_VECTOR_LONG_INT_H


/** @name Real-valued matrices */
//@{
 /** Compute the solution to a real system of linear equations A*X=B
 *
 * Depending on the dimensions of A, either a LU or a QR decomposition
 * is used.
 */
void LaLinearSolve( const LaGenMatDouble& A, LaGenMatDouble& X, 
		    const LaGenMatDouble& B );

/** Compute the solution to a real system of linear equations A*X=B
 * in-place.
 *
 * Depending on the dimensions of A, either a LU or a QR decomposition
 * is used.
 *
 * In-place means: The contents of A are overwritten during the
 * calculation. B is not overwritten but always copied during this
 * operation.
 */
void LaLinearSolveIP( LaGenMatDouble& A, LaGenMatDouble& X, 
		      const LaGenMatDouble& B );


/** Compute the solution to a real system of linear equations A*X=B by
 * using the LU decomposition. This only works for a squares matrix A.
 */
void LaLULinearSolve( const LaGenMatDouble& A, LaGenMatDouble& X, 
		      const LaGenMatDouble& B );

/** Compute the solution to a real system of linear equations A*X=B by
 * using the LU decomposition in-place. This only works for a squares
 * matrix A.
 *
 * In-place means: The contents of A are overwritten during the
 * calculation. B is not overwritten but always copied during this
 * operation.
 */
void LaLULinearSolveIP( LaGenMatDouble& A, LaGenMatDouble& X, 
			const LaGenMatDouble& B );

/** Compute the solution to a real system of linear equations A*X=B by
 * using QR decomposition, which works for any rectangular matrix A.
 */
void LaQRLinearSolve( const LaGenMatDouble& A, LaGenMatDouble& X, 
		      const LaGenMatDouble& B );

/** Compute the solution to a real system of linear equations A*X=B by
 * using the QR decomposition in-place. This works for any rectangular 
 * matrix A.
 *
 * In-place means: The contents of A are overwritten during the
 * calculation. B is not overwritten but always copied during this
 * operation.
 */
void LaQRLinearSolveIP( LaGenMatDouble& A, LaGenMatDouble& X, 
			const LaGenMatDouble& B );

/** @brief Compute the LU factorization
 *
 * Compute the LU factorization (in-place) of a general M-by-N
 * matrix A.
 *
 * More info: See man @c dgetrf. 
 *
 * @param GM Matrix to be factorized in-place.
 * @param PIV Vector to return the pivoting indices. 
 * This vector *has* to be at least as long as min(M,N).
 */
void LUFactorizeIP(LaGenMatDouble &GM, LaVectorLongInt &PIV);
//@}



#ifdef LA_COMPLEX_SUPPORT
/** @name Complex-valued matrices */
//@{
 /** Compute the solution to a complex-valued system of linear
 * equations A*X=B
 *
 * Depending on the dimensions of A, either a LU or a QR decomposition
 * is used.
 */
void LaLinearSolve( const LaGenMatComplex& A, LaGenMatComplex& X, 
		    const LaGenMatComplex& B );

/** Compute the solution to a complex-valued system of linear
 * equations A*X=B in-place.
 *
 * Depending on the dimensions of A, either a LU or a QR decomposition
 * is used.
 *
 * In-place means: The contents of A are overwritten during the
 * calculation. B is not overwritten but always copied during this
 * operation.
 */
void LaLinearSolveIP( LaGenMatComplex& A, LaGenMatComplex& X, 
		      const LaGenMatComplex& B );


/** Compute the solution to a complex-valued system of linear
 * equations A*X=B by using the LU decomposition. This only works for
 * a squares matrix A.
 */
void LaLULinearSolve( const LaGenMatComplex& A, LaGenMatComplex& X, 
		      const LaGenMatComplex& B );

/** Compute the solution to a complex-valued system of linear
 * equations A*X=B by using the LU decomposition in-place. This only
 * works for a squares matrix A.
 *
 * In-place means: The contents of A are overwritten during the
 * calculation. B is not overwritten but always copied during this
 * operation.
 */
void LaLULinearSolveIP( LaGenMatComplex& A, LaGenMatComplex& X, 
			const LaGenMatComplex& B );

/** Compute the solution to a complex-valued system of linear
 * equations A*X=B by using QR decomposition, which works for any
 * rectangular matrix A.
 */
void LaQRLinearSolve( const LaGenMatComplex& A, LaGenMatComplex& X, 
		      const LaGenMatComplex& B );

/** Compute the solution to a complex-valued system of linear
 * equations A*X=B by using the QR decomposition in-place. This works
 * for any rectangular matrix A.
 *
 * In-place means: The contents of A are overwritten during the
 * calculation. B is not overwritten but always copied during this
 * operation.
 */
void LaQRLinearSolveIP( LaGenMatComplex& A, LaGenMatComplex& X, 
			const LaGenMatComplex& B );

/** @brief Compute the LU factorization
 *
 * Compute the LU factorization (in-place) of a general M-by-N
 * matrix A.
 *
 * More info: See man @c zgetrf. 
 *
 * @param GM Matrix to be factorized in-place.
 * @param PIV Vector to return the pivoting indices. 
 * This vector *has* to be at least as long as min(M,N).
 */
void LUFactorizeIP(LaGenMatComplex &GM, LaVectorLongInt &PIV);
//@}
#endif // LA_COMPLEX_SUPPORT


#ifdef _LA_SPD_MAT_DOUBLE_H_

void LaLinearSolve( const LaSpdMatDouble& A, LaGenMatDouble& X, 
    LaGenMatDouble& B );
void LaLinearSolveIP( LaSpdMatDouble& A, LaGenMatDouble& X, LaGenMatDouble& B );

void LaCholLinearSolve( const LaSpdMatDouble& A, LaGenMatDouble& X, 
        LaGenMatDouble& B );
void LaCholLinearSolveIP( LaSpdMatDouble& A, LaGenMatDouble& X, 
        LaGenMatDouble& B );
#endif // _LA_SPD_MAT_DOUBLE_H_

#ifdef _LA_SYMM_MAT_DOUBLE_H_

/** FIXME: Document me! FIXME: Needs verification. */
void LaLinearSolve( const LaSymmMatDouble& A, LaGenMatDouble& X, 
		    const LaGenMatDouble& B );
/** FIXME: Document me! FIXME: Needs verification. */
void LaLinearSolveIP( LaSymmMatDouble& A, LaGenMatDouble& X, 
		      const LaGenMatDouble& B );

/** FIXME: Document me! FIXME: Needs verification. */
void LaCholLinearSolve( const LaSymmMatDouble& A, LaGenMatDouble& X, 
			const LaGenMatDouble& B );
/** FIXME: Document me! FIXME: Needs verification. */
void LaCholLinearSolveIP( LaSymmMatDouble& A, LaGenMatDouble& X, 
			  const LaGenMatDouble& B );

// Eigenvalue problems 

void LaEigSolve(const LaSymmMatDouble &S, LaVectorDouble &eigvals);
void LaEigSolve(const LaSymmMatDouble &S, LaVectorDouble &eigvals, 
    LaGenMatDouble &eigvec);
void LaEigSolveIP(LaSymmMatDouble &S, LaVectorDouble &eigvals);
#endif // _LA_SYMM_MAT_DOUBLE_H_

/** FIXME: This is a misleading function! This function calculates all
 * eigenvalues and eigenvectors of a <i>symmetric</i> matrix A, <i>not
 * a general matrix A</i>!
 *
 * @param A On entry, the symmetric (not a general) matrix A. The
 * leading N-by-N lower triangular part of A is used as the lower
 * triangular part of the matrix to be decomposed. On exit, A contains
 * the orthonormal eigenvectors of the matrix.
 *
 * @param eigvals On exit, this vector contains the eigenvalues in ascending order.
 *
 * FIXME: Needs verification! FIXME: This uses dsyev which only works
 * on symmetric matrices. Instead, this should be changed to use dgeev
 * or even better dgeevx. For the complex case, we would have to write
 * another function that uses zgeev or zgeevx. */
void LaEigSolveVecIP(LaGenMatDouble &A, LaVectorDouble &eigvals);



#endif // _LASLV_H
