//
//              LAPACK++ 1.1 Linear Algebra Package 1.1
//               University of Tennessee, Knoxvilee, TN.
//            Oak Ridge National Laboratory, Oak Ridge, TN.
//        Authors: J. J. Dongarra, E. Greaser, R. Pozo, D. Walker
//                 (C) 1992-1996 All Rights Reserved
//
//                             NOTICE
//
// Permission to use, copy, modify, and distribute this software and
// its documentation for any purpose and without fee is hereby granted
// provided that the above copyright notice appear in all copies and
// that both the copyright notice and this permission notice appear in
// supporting documentation.
//
// Neither the Institutions (University of Tennessee, and Oak Ridge National
// Laboratory) nor the Authors make any representations about the suitability 
// of this software for any purpose.  This software is provided ``as is'' 
// without express or implied warranty.
//
// LAPACK++ was funded in part by the U.S. Department of Energy, the
// National Science Foundation and the State of Tennessee.

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "arch.h"
#include "lafnames.h"
#include LA_PREFS_H
#include LA_GEN_MAT_COMPLEX_H

DLLIMPORT int LaGenMatComplex::debug_ = 0;     // turn off global deubg flag initially.
                                // use A.debug(1) to turn on/off,
                                // and A.debug() to check current status.

DLLIMPORT int* LaGenMatComplex::info_= new int;        // turn off info print flag.

LaGenMatComplex::~LaGenMatComplex()
{}

LaGenMatComplex::LaGenMatComplex() : v(0)
{

    dim[0] = dim[1] = 0;
    size0 = size1 = 0; //sz[0] = sz[1] = 0;
    *info_ = 0;
    shallow_= 0;
}



LaGenMatComplex::LaGenMatComplex(int m, int n) :v(m*n)
{
    ii[0](0,m-1);
    ii[1](0,n-1);
    dim[0] = m;
    dim[1] = n;
    size0 = m;
    size1 = n;
    *info_ = 0;
    shallow_= 0;
}

LaGenMatComplex::LaGenMatComplex(COMPLEX *d, int m, int n, bool row_ordering) :v(d, m, n, row_ordering)
{
    if (debug())
    {
        std::cout << ">>> LaGenMatComplex::LaGenMatComplex(COMPLEX *d, int m, int n, bool row_ordering) :v(d, m, n, row_ordering)\n";
    }

    ii[0](0,m-1);
    ii[1](0,n-1);
    dim[0] = m;
    dim[1] = n;
    size0 = m;
    size1 = n;
    *info_ = 0;
    shallow_= 0;  

    if (debug())
    {
        std::cout << "<<< LaGenMatComplex::LaGenMatComplex(COMPLEX *d, int m, int n, bool row_ordering) :v(d, m, n, row_ordering)\n";
    }
}

LaGenMatComplex& LaGenMatComplex::operator=(COMPLEX s)
{
   if (debug())
   {
      std::cout << ">>> LaGenMatComplex& LaGenMatComplex::operator=(COMPLEX s)\n";
      std::cout << "       s = " << s << std::endl;
   }

   v = s;
    
   if (debug())
   {
      std::cout << "  *this: " << this->info() << std::endl;
      std::cout << " >>> LaGenMatComplex& LaGenMatComplex::operator=(COMPLEX s)\n";
   }
    
   return *this;
}

//LaGenMatComplex& LaGenMatComplex::operator=(const std::complex<double>& s)
//{
//    return operator=(LaComplex(s));
//}

LaGenMatComplex& LaGenMatComplex::operator=(const LaComplex& s)
{
    return operator=(s.toCOMPLEX());
}

LaGenMatComplex& LaGenMatComplex::ref(const LaGenMatComplex& s)
{

   // handle trivial M.ref(M) case
   if (this == &s) return *this;
   else
   {
      ii[0] = s.ii[0];
      ii[1] = s.ii[1];
      dim[0] = s.dim[0];
      dim[1] = s.dim[1];
      size0 = s.size0;
      size1 = s.size1;
      shallow_ = 0;

      v.ref(s.v);

      return *this;
   }
}

LaGenMatComplex&  LaGenMatComplex::resize(int m, int n)
{
    if (debug())
    {
        std::cout << ">>> resize("<< m << "," << n << ")\n";
    }

    // first, reference 0x0 matrix, potentially freeing memory
    // this allows one to resize a matrix > 1/2 of the available
    // memory

    LaGenMatComplex tmp1(0,0);
    ref(tmp1);

    // now, reference an MxN matrix
    LaGenMatComplex tmp(m,n);
    ref(tmp);

    if (debug())
    {
       std::cout << "<<< resize: *this: " << this->info() << std::endl;
    }
    return *this;

}

LaGenMatComplex&  LaGenMatComplex::resize(const LaGenMatComplex &A)
{
   if (debug())
   {
      std::cout << ">>> resize(const LaGenMatComplex &A)" << std::endl;
      std::cout << "      A: " << A.info() << std::endl;
   }

   resize(A.size0, A.size1);
   if (debug())
   {
      std::cout << "<<< resize(const LaGenMatComplex &A) " << std::endl;
   }
   return *this;
}


LaGenMatComplex::LaGenMatComplex(const LaGenMatComplex& X) : v(0)
{
   if (X.debug())
   {
      std::cout << ">>> LaGenMatComplex::LaGenMatComplex(const LaGenMatComplex& X)\n";
      std::cout << "X: " << X.info() << std::endl;

   }
   debug_ = X.debug_;

   shallow_ = 0;  // do not perpeturate shallow copies, otherwise
   //  B = A(I,J) does not work properly...

   if (X.shallow_)
   {
      v.ref(X.v);
      dim[0] = X.dim[0];
      dim[1] = X.dim[1];
      size0 = X.size0;
      size1 = X.size1;
      ii[0] = X.ii[0];
      ii[1] = X.ii[1];
   }
   else
   {
      if (X.debug())
	 std::cout << std::endl << "Data is being copied!\n" << std::endl;
      v.resize(X.size0*X.size1); 
      ii[0](0,X.size0-1);
      ii[1](0,X.size1-1);
      dim[0] = size0 = X.size0;
      dim[1] = size1 = X.size1;
      if (inc(0) == 1 && inc(1) == 1)
	 // Unit strides
	 v.copy(X.v);
      else
      {
	 int i, j, M = X.size0, N = X.size1;
	 for (j=0; j<N; j++)
	    for (i=0; i<M; i++)
	       (*this)(i,j) = X(i,j);
      }
   }  

   if (debug())
   {
      std::cout << "*this: " << info() << std::endl;
      std::cout << "<<< LaGenMatComplex::LaGenMatComplex(const LaGenMatComplex& X)\n";
   }
}


LaGenMatComplex& LaGenMatComplex::copy(const LaGenMatComplex& X) 
{
   if (debug())
   {
      std::cout << ">>> LaGenMatComplex& LaGenMatComplex::copy(const LaGenMatComplex& X)\n";
      std::cout << " X: " << X.info() << std::endl;
   }

   // current scheme in copy() is to detach the left-hand-side
   // from whatever it was pointing to.
   //
   resize(X);

   if (inc(0) == 1 && inc(1) == 1)
      // Unit strides
      v.copy(X.v);
   else
   {
      int i,j, M = X.size0,  N = X.size1;
      for (i=0; i<M; i++)
	 for (j=0; j<N; j++)
	    (*this)(i,j) = X(i,j);
   }

   if ((debug()))
   {
      std::cout << " *this: " << this->info() << std::endl;
      std::cout << " <<< LaGenMatComplex& LaGenMatComplex::copy \
(const LaGenMatComplex& s)\n";
   }

   return *this;
}


LaGenMatComplex& LaGenMatComplex::inject(const LaGenMatComplex& s)
{
    assert(s.size0 == size0);
    assert(s.size1 == size1);

    if (inc(0) == 1 && inc(1) == 1)
       // Unit strides
       v.inject(s.v);
    else
    {
       int i, j,  M=size0, N=size1;
       for (j=0;j<N;j++)
	  for (i=0;i<M; i++)
	     operator()(i,j) = s(i,j);
    }

    return *this;
}


LaGenMatComplex LaGenMatComplex::operator()(const LaIndex& II, const LaIndex& JJ) const
{
   if (debug())
   {
      std::cout << ">>> LaGenMatComplex::operator(const LaIndex& const LaIndex&)\n";
   }
   LaIndex I, J;

    if (II.null())
        I(0, size0-1);
    else
        I = II;
    if (JJ.null())
        J(0, size1-1);
    else
        J = JJ;

    //(JJ.null()) ?  J(0,size1) : J = JJ;

    assert(I.inc() != 0);
    assert(J.inc() != 0);

    if (I.inc() > 0)
    {
        assert(I.start() >= 0);
        assert(I.start() <= I.end());
        assert(I.end() < size0);
    }
    else // I.inc() < 0
    {
        assert(I.start() < size0);
        assert(I.start() >= I.end());
        assert(I.end() >= 0);
    }

    if (J.inc() > 0)
    {
        assert(J.start() >= 0);
        assert(J.start() <= J.end());
        assert(J.end() < size1);
    }
    else  // J.inc() < 0
    {
        assert(J.start() < size1);
        assert(J.start() >= J.end());
        assert(J.end() >= 0);
    }

    LaGenMatComplex tmp;

    tmp.dim[0] = dim[0];
    tmp.dim[1] = dim[1];
    tmp.size0 = (I.end() - I.start())/I.inc() + 1;
    tmp.size1 = (J.end() - J.start())/J.inc() + 1;

    tmp.ii[0].start() =  ii[0].start() + I.start()*ii[0].inc();
    tmp.ii[0].inc() = ii[0].inc() * I.inc();
    tmp.ii[0].end() = (I.end() - I.start())/ I.inc() * tmp.ii[0].inc() 
            + tmp.ii[0].start();

    tmp.ii[1].start() =  ii[1].start() + J.start()*ii[1].inc();
    tmp.ii[1].inc() = ii[1].inc() * J.inc();
    tmp.ii[1].end() = (J.end() - J.start())/ J.inc() * tmp.ii[1].inc() 
            + tmp.ii[1].start();

    tmp.v.ref(v);
    tmp.shallow_assign();

    if (debug())
    {
       std::cout << "    return value: " << tmp.info() << std::endl;
       std::cout << "<<< LaGenMatComplex::operator(const LaIndex& const LaIndex&)\n";
    }

    return tmp;

}

LaGenMatComplex LaGenMatComplex::operator()(const LaIndex& II, const LaIndex& JJ) 
{
   if (debug())
   {
      std::cout << ">>> LaGenMatComplex::operator(const LaIndex& const LaIndex&)\n";
   }
   LaIndex I, J;

    if (II.null())
        I(0,size0-1);
    else
        I = II;
    if (JJ.null())
        J(0,size1-1);
    else
        J = JJ;

    //(JJ.null()) ?  J(0,size1) : J = JJ;

    assert(I.inc() != 0);
    assert(J.inc() != 0);

    if (I.inc() > 0)
    {
        assert(I.start() >= 0);
        assert(I.start() <= I.end());
        assert(I.end() < size0);
    }
    else // I.inc() < 0
    {
        assert(I.start() < size0);
        assert(I.start() >= I.end());
        assert(I.end() >= 0);
    }

    if (J.inc() > 0)
    {
        assert(J.start() >= 0);
        assert(J.start() <= J.end());
        assert(J.end() < size1);
    }
    else  // J.inc() < 0
    {
        assert(J.start() < size1);
        assert(J.start() >= J.end());
        assert(J.end() >= 0);
    }

    LaGenMatComplex tmp;

    tmp.dim[0] = dim[0];
    tmp.dim[1] = dim[1];
    tmp.size0 = (I.end() - I.start())/I.inc() + 1;
    tmp.size1 = (J.end() - J.start())/J.inc() + 1;

    tmp.ii[0].start() =  ii[0].start() + I.start()*ii[0].inc();
    tmp.ii[0].inc() = ii[0].inc() * I.inc();
    tmp.ii[0].end() = (I.end() - I.start())/ I.inc() * tmp.ii[0].inc() 
            + tmp.ii[0].start();

    tmp.ii[1].start() =  ii[1].start() + J.start()*ii[1].inc();
    tmp.ii[1].inc() = ii[1].inc() * J.inc();
    tmp.ii[1].end() = (J.end() - J.start())/ J.inc() * tmp.ii[1].inc() 
            + tmp.ii[1].start();

    tmp.v.ref(v);
    tmp.shallow_assign();

    if (debug())
    {
       std::cout << "    return value: " << tmp.info() << std::endl;
       std::cout << "<<< LaGenMatComplex::operator(const LaIndex& const LaIndex&)\n";
    }

    return tmp;

}


std::ostream& operator<<(std::ostream& s, const LaGenMatComplex& G)
{
    if (*(G.info_))     // print out only matrix info, not actual values
    {
        *(G.info_) = 0; // reset the flag
        s << "(" << G.size0 << "x" << G.size1 << ") " ;
        s << "Indices: " << G.ii[0] << " " << G.ii[1];
        s << " #ref: " << G.ref_count();
        s << " shallow:" << G.shallow_  ;
    }
    else 
    {

        int i,j;
        LaPreferences::pFormat p = LaPreferences::getPrintFormat();
        bool newlines = LaPreferences::getPrintNewLines();

        if((p == LaPreferences::MATLAB) || (p == LaPreferences::MAPLE))
          s << "[";

        for (i=0; i<G.size0; i++)
        {
            if(p == LaPreferences::MAPLE)
              s << "[";
            for (j=0; j<G.size1; j++)
            {
              if(p == LaPreferences::NORMAL)
                s << G(i,j);
              if(p == LaPreferences::MATLAB)
                s << G(i,j).r << "+" << G(i,j).i << "i";
              if(p == LaPreferences::MAPLE)
                s << G(i,j).r << "+" << G(i,j).i << "*I";
              if(((p == LaPreferences::NORMAL) || (p == LaPreferences::MATLAB)) && (j != G.size(1)-1))
                s << "  ";
              if(((p == LaPreferences::MAPLE)) && (j != G.size(1)-1))
                s << ", ";              
            }
            if(p == LaPreferences::MAPLE)
            {
              s << "]";
              if(i != G.size(0)-1)
                s << ", ";
            }
            if((p == LaPreferences::MATLAB) && (i != G.size(0)-1))
              s << ";  ";
            if( ((newlines)||(p==LaPreferences::NORMAL)) && (i != G.size(0)-1)) // always print newline if in NORMAL mode
              s << "\n";
        }
        if((p == LaPreferences::MATLAB) || (p == LaPreferences::MAPLE))
          s << "]";

        s << "\n";
    }
    return s;
}

