//
//              LAPACK++ 1.1 Linear Algebra Package 1.1
//               University of Tennessee, Knoxvilee, TN.
//            Oak Ridge National Laboratory, Oak Ridge, TN.
//        Authors: J. J. Dongarra, E. Greaser, R. Pozo, D. Walker
//                 (C) 1992-1996 All Rights Reserved
//
//                             NOTICE
//
// Permission to use, copy, modify, and distribute this software and
// its documentation for any purpose and without fee is hereby granted
// provided that the above copyright notice appear in all copies and
// that both the copyright notice and this permission notice appear in
// supporting documentation.
//
// Neither the Institutions (University of Tennessee, and Oak Ridge National
// Laboratory) nor the Authors make any representations about the suitability 
// of this software for any purpose.  This software is provided ``as is'' 
// without express or implied warranty.
//
// LAPACK++ was funded in part by the U.S. Department of Energy, the
// National Science Foundation and the State of Tennessee.


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "lafnames.h"
#include VECTOR_COMPLEX_H
#include LA_EXCEPTION_H

VectorComplex::VectorComplex(int n=0)
{                                                                      
    assert(n>=0);
    p = new vrefComplex;
    p->sz = n;                                                          
    p->data = data = new COMPLEX[n]; 
    p->ref_count = 1;                        
}                                                                      


VectorComplex::VectorComplex(const VectorComplex& m)
{
   // shallow assignment semantics

    p = m.p;
    data = m.data;
    p->ref_count++;
}

VectorComplex::VectorComplex(COMPLEX *d, int n)
{                                                                      
    p = new vrefComplex;
    p->sz = n;                                                          
    p->ref_count = 2; 
    p->data = data = d;                        
}                                                                      
                                                                       
VectorComplex::VectorComplex(COMPLEX *d, int m, int n, bool row_ordering)
{
    p = new vrefComplex;
    p->sz = m*n;
    if(!row_ordering)
    {
      p->ref_count = 2;
      p-> data = data = d;
    }
    else // row ordering
    {
      int i;
      if (d == NULL)
        throw LaException("VectorComplex(*d, m, n, ordering)", "data is NULL");
      p->ref_count = 1;
      p->data = data = new COMPLEX[m*n];
      for(i=0; i < m*n; i++)
      {
        data[m*(i%n)+(i/n)]=d[i];  // reorder the data to column-major
      }
    }
}
                                                                       

VectorComplex::~VectorComplex()
{
   if (--(p->ref_count) == 0)              // perform garbage col.
   {
      delete [] p->data;
      delete p;
   }
}

VectorComplex::VectorComplex(int n, COMPLEX scalar)
{                                                                      
    p = new vrefComplex;
    p->sz = n;                                                          
    p->ref_count = 1;                        
    p->data = data = new COMPLEX[n]; 

    double s_re = scalar.r;
    double s_im = scalar.i;
    COMPLEX * iter = data;
    COMPLEX * end = data + n;
    for ( ; iter != end; ++iter)
    {
       iter->r = s_re;
       iter->i = s_im;
    }

    //for (int i=0; i<n; i++)                                            
    //  data[i] = scalar;
}                                                                      
                                                                       


// this actually frees memory first, then resizes it.  it reduces
// internal fragmentation of memory pool, and the resizing of
// matrices > 1/2 available memory.

int VectorComplex::resize(int d)
{

    if (d<0)                // do nothing if invalid size
    {
        return size();
    }
    else
    {
        ref(VectorComplex(0));  // possibilby free up destination
        if (d>0)
            ref(VectorComplex(d));
    }
    return size();
}


VectorComplex& VectorComplex::inject(const VectorComplex& m)
{
    if (m.size() != size())
    {
       throw LaException("VectorComplex::inject()", "Vector sizes do not match.\n");
    }
    int N = size();

    COMPLEX *dest = data;
    COMPLEX *src = m.data;
//     for (int i=0; i<N; i++)
//     {
//        dest[i].r = src[i].r;
//        dest[i].i = src[i].i;
//     }
    memcpy(dest, src, N*sizeof(COMPLEX));

    return *this;
}


VectorComplex& VectorComplex::copy(const VectorComplex &m)
{
    resize(0);                  // free up destination

    int N = m.size();
    VectorComplex tmp(N);

    COMPLEX *dest = tmp.data;
    COMPLEX *src = m.data;
//     for (int i=0; i<N; i++)
//     {
//        dest[i].r = src[i].r;
//        dest[i].i = src[i].i;
//     }
    memcpy(dest, src, N*sizeof(COMPLEX));

    ref(tmp);
    return *this;
}


std::ostream& operator<<(std::ostream& s, const VectorComplex& m)
{
   if (m.p)
   {
      int n = m.size();
      for (int i=0; i<n; i++)
	 s << m(i) << "  ";
      s << "\n";
   }
   else s << "NULL VectorComplex.\n";
   return s;
}



VectorComplex& VectorComplex::operator=(COMPLEX scalar)
{
   //for (int i=0; i<size(); i++)
   //      data[i] = scalar;

   // This function has been heavily optimized.

   // Cache the complex value
   double s_re = scalar.r;
   double s_im = scalar.i;
   COMPLEX * iter = data;
   COMPLEX * end;
   const int _blocksize = 8;

   // This idea and algorithm is borrowed from dscal.f
   int m = size() % _blocksize;
   if (m != 0) 
   {
      end = data + m;
      for ( ; iter != end; ++iter)
      {
	 iter->r = s_re;
	 iter->i = s_im;
      }
      if (size() < _blocksize)
	 return *this;
   }

   end = data + size();
   for ( ; iter != end; iter += _blocksize)
   {
      iter->r = 
	 iter[1].r = 
	 iter[2].r = 
	 iter[3].r = 
	 iter[4].r = 
	 iter[5].r = 
	 iter[6].r = 
	 iter[7].r = 
	 s_re;
      iter->i =
	 iter[1].i =
	 iter[2].i =
	 iter[3].i =
	 iter[4].i =
	 iter[5].i =
	 iter[6].i =
	 iter[7].i =
	 s_im;
   }

   return *this;
}
